package net.hostsharing.hsadminng.hs.hosting.asset;

import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import net.hostsharing.hsadminng.hs.booking.generated.api.v1.model.HsHostingAssetAutoInsertResource;
import net.hostsharing.hsadminng.hs.booking.generated.api.v1.model.HsHostingAssetTypeResource;
import net.hostsharing.hsadminng.hs.booking.item.BookingItemCreatedEvent;
import net.hostsharing.hsadminng.hs.booking.item.BookingItemCreatedEvent.Status;
import net.hostsharing.hsadminng.hs.booking.item.HsBookingItemRealEntity;
import net.hostsharing.hsadminng.hs.hosting.asset.validators.HostingAssetEntitySaveProcessor;
import net.hostsharing.hsadminng.hs.office.contact.HsOfficeContactRealEntity;
import net.hostsharing.hsadminng.mapper.StandardMapper;
import net.hostsharing.hsadminng.persistence.EntityManagerWrapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationListener;
import org.springframework.stereotype.Component;

import jakarta.validation.ValidationException;
import java.net.IDN;
import java.util.Map;
import java.util.UUID;

import static java.util.Map.entry;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.DOMAIN_DNS_SETUP;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.DOMAIN_HTTP_SETUP;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.DOMAIN_MBOX_SETUP;
import static net.hostsharing.hsadminng.hs.hosting.asset.HsHostingAssetType.DOMAIN_SMTP_SETUP;

@Component
public class HsBookingItemCreatedListener implements ApplicationListener<BookingItemCreatedEvent> {

    @Autowired
    private EntityManagerWrapper emw;

    @Autowired
    private ObjectMapper jsonMapper;

    @Autowired
    private StandardMapper standardMapper;

    @Override
    @SneakyThrows
    public void onApplicationEvent(final BookingItemCreatedEvent event) {
        final var newBookingItemRealEntity =
                emw.getReference(HsBookingItemRealEntity.class, event.getBookingItemUuid());
        final var asset = jsonMapper.readValue(event.getAssetJson(), HsHostingAssetAutoInsertResource.class);
        final var factory = switch (newBookingItemRealEntity.getType()) {
            case PRIVATE_CLOUD -> null;
            case CLOUD_SERVER -> null;
            case MANAGED_SERVER -> null;
            case MANAGED_WEBSPACE -> null;
            case DOMAIN_SETUP -> new DomainSetupHostingAssetFactory(newBookingItemRealEntity, asset);
        };
        if (factory != null) {
            event.setStatus(factory.performSaveProcess());
        }
    }

    private <T> T ref(final Class<T> entityClass, final UUID uuid) {
        return uuid != null ? emw.getReference(entityClass, uuid) : null;
    }

    @RequiredArgsConstructor
    abstract class HostingAssetFactory {

        final HsBookingItemRealEntity fromBookingItem;
        final HsHostingAssetAutoInsertResource asset;

        protected HsHostingAsset create() {
            return null;
        }

        Status performSaveProcess() {
            final var newHostingAsset = create();
            try {
                return persist(newHostingAsset);
            } catch (final Exception e) {
                return Status.failed(e.getMessage());
            }
        }

        protected Status persist(final HsHostingAsset newHostingAsset) {
            new HostingAssetEntitySaveProcessor(emw, newHostingAsset)
                    .preprocessEntity()
                    .validateEntity()
                    .prepareForSave()
                    .save()
                    .validateContext();
            return Status.finished();
        }
    }

    class DomainSetupHostingAssetFactory extends HostingAssetFactory {

        public DomainSetupHostingAssetFactory(
                final HsBookingItemRealEntity newBookingItemRealEntity,
                final HsHostingAssetAutoInsertResource asset) {
            super(newBookingItemRealEntity, asset);
        }

        @Override
        protected HsHostingAsset create() {
            final String domainName = asset.getIdentifier();
            final var domainSetupAsset = createDomainSetupAsset(domainName);

            // TODO.legacy: as long as we need to be compatible, we always do all technical domain-setups
            final var subHostingAssetResources = asset.getSubHostingAssets();
            final var domainHttpSetupAssetResource = subHostingAssetResources.stream()
                    .filter(ha -> ha.getType() == HsHostingAssetTypeResource.DOMAIN_HTTP_SETUP)
                    .findFirst().orElseThrow(() -> new ValidationException(
                            domainName + ": missing target unix user (assignedToHostingAssetUuid) for DOMAIN_HTTP_SETUP "));
            final var domainHttpSetupAsset = domainSetupAsset.getSubHostingAssets().stream().filter(sha -> sha.getType() == DOMAIN_HTTP_SETUP).findFirst().orElseThrow();
            domainHttpSetupAsset.setParentAsset(domainSetupAsset);
            final HsHostingAssetRealEntity assignedToUnixUserAsset =
                    emw.find(HsHostingAssetRealEntity.class, domainHttpSetupAssetResource.getAssignedToAssetUuid());
            domainHttpSetupAsset.setAssignedToAsset(assignedToUnixUserAsset);

            if (subHostingAssetResources.stream().noneMatch(ha -> ha.getType() == HsHostingAssetTypeResource.DOMAIN_DNS_SETUP)) {
                domainSetupAsset.getSubHostingAssets().add(HsHostingAssetRealEntity.builder()
                        .type(DOMAIN_DNS_SETUP)
                        .parentAsset(domainSetupAsset)
                        .assignedToAsset(assignedToUnixUserAsset.getParentAsset()) // FIXME: why is that needed?
                        .identifier(domainName + "|DNS")
                        .config(Map.ofEntries(
                                // FIXME:
                                entry("TTL", 21600),
                                entry("auto-SOA", true)
                        ))
                        .build());
            }
            if (subHostingAssetResources.stream().noneMatch(ha -> ha.getType() == HsHostingAssetTypeResource.DOMAIN_MBOX_SETUP)) {
                domainSetupAsset.getSubHostingAssets().add(HsHostingAssetRealEntity.builder()
                        .type(DOMAIN_MBOX_SETUP)
                        .parentAsset(domainSetupAsset)
                        .assignedToAsset(assignedToUnixUserAsset.getParentAsset())
                        .identifier(domainName + "|MBOX")
                        .caption("HTTP-Setup für " + IDN.toUnicode(domainName))
                        .build());
            }
            if (subHostingAssetResources.stream().noneMatch(ha -> ha.getType() == HsHostingAssetTypeResource.DOMAIN_SMTP_SETUP)) {
                domainSetupAsset.getSubHostingAssets().add(HsHostingAssetRealEntity.builder()
                        .type(DOMAIN_SMTP_SETUP)
                        .parentAsset(domainSetupAsset)
                        .assignedToAsset(assignedToUnixUserAsset.getParentAsset())
                        .identifier(domainName + "|SMTP")
                        .caption("HTTP-Setup für " + IDN.toUnicode(domainName))
                        .build());
            }
            return domainSetupAsset;
        }

        private HsHostingAssetRealEntity createDomainSetupAsset(final String domainName) {
            return HsHostingAssetRealEntity.builder()
                    .bookingItem(fromBookingItem)
                    .type(HsHostingAssetType.DOMAIN_SETUP)
                    .identifier(domainName)
                    .caption(asset.getCaption() != null ? asset.getCaption() : domainName)
                    .parentAsset(ref(HsHostingAssetRealEntity.class, asset.getParentAssetUuid()))
                    .alarmContact(ref(HsOfficeContactRealEntity.class, asset.getAlarmContactUuid()))
                    // FIXME .config(asset.getConfig())
                    .subHostingAssets(
                            standardMapper.mapList(asset.getSubHostingAssets(), HsHostingAssetRealEntity.class)
                    )
                    .build();
        }

        @Override
        protected Status persist(final HsHostingAsset newHostingAsset) {
            final var status = super.persist(newHostingAsset);
            newHostingAsset.getSubHostingAssets().forEach(super::persist);
            return status;
        }
    }
}
